"""Interfaces and classes for Moving Targets Masters."""

from moving_targets.masters.classification_master import ClassificationMaster
from moving_targets.masters.master import Master
from moving_targets.masters.regression_master import RegressionMaster
